from dune.perftool.ufl.modified_terminals import Restriction
from dune.perftool.generation import (iname,
                                      domain,
                                      get_global_context_value,
                                      temporary_variable,
                                      instruction)
from dune.perftool.pdelab.geometry import (local_dimension,
                                           world_dimension,
                                           name_localcenter,
                                           pymbolic_in_cell_coordinates)

from dune.perftool.pdelab.quadrature import (quadrature_inames)
from dune.perftool.generation.counter import get_counted_variable
from dune.perftool.options import get_option
import pymbolic.primitives as prim


@iname
def sub_element_inames():
    name = "subel"
    dim = local_dimension()
    inames = tuple()
    for i in range(dim):
        name_counted = get_counted_variable(name)
        inames = inames + (name_counted,)
        domain(name_counted, get_option("number_of_blocks"))
    return inames


def sub_facet_inames():
    subelem_inames = sub_element_inames()
    
    center = pymbolic_in_cell_coordinates(prim.Variable(name_localcenter()), Restriction.NEGATIVE)

    def predicate(index):
        return prim.Comparison(prim.Subscript(center, index), "==", 0.5)

    def conditional_instruction(index):
        instruction(assignee=prim.Variable(inames[index]),
                    expression=prim.Variable(subelem_inames[1 if index==2 else 0]),
                    within_inames=frozenset(subelem_inames).union(frozenset(quadrature_inames())),
                    predicates=frozenset([predicate(index)])
                    )
        instruction(assignee=prim.Variable(inames[index]),
                    expression=prim.Product(((k-1),prim.Subscript(center, (index,)))),
                    within_inames=frozenset(subelem_inames).union(frozenset(quadrature_inames())),
                    predicates=frozenset([prim.LogicalNot(predicate(index))])
                    )

    k = get_option("number_of_blocks")

    inames = ("x",)
    temporary_variable(inames[0])
    conditional_instruction(0)
    if world_dimension() < 3:
        inames = inames + ("y",)
        temporary_variable(inames[1])
        conditional_instruction(1)
    else:
        inames = inames + ("y",)
        temporary_variable(inames[1])
        instruction(assignee=prim.Variable(inames[1]),
                    expression=prim.Variable(subelem_inames[0]),
                    within_inames=frozenset(subelem_inames).union(frozenset(quadrature_inames())),
                    predicates=frozenset([prim.LogicalAnd((predicate(1), predicate(2)))])
                    )
        instruction(assignee=prim.Variable(inames[1]),
                    expression=prim.Variable(subelem_inames[1]),
                    within_inames=frozenset(subelem_inames).union(frozenset(quadrature_inames())),
                    predicates=frozenset([prim.LogicalAnd((predicate(1), predicate(0)))])
                    )
        instruction(assignee=prim.Variable(inames[1]),
                    expression=prim.Product(((k-1),prim.Subscript(center, (1,)))),
                    within_inames=frozenset(subelem_inames).union(frozenset(quadrature_inames())),
                    predicates=frozenset([prim.LogicalNot(predicate(1))])
                    )

        inames = inames + ("z",)
        temporary_variable(inames[2])
        conditional_instruction(2)

    return inames


def tensor_index_to_sequential_index(indices, k):
    return prim.Sum(tuple(index * k ** i for i, index in enumerate(indices)))


# TODO better name
def to_tensor_index(iname):
    return tuple(prim.RightShift(prim.BitwiseAnd((prim.Variable(iname),2**n)), n)  for n in range(world_dimension()))


# TODO 3d
def micro_index_to_macro_index(iname):
    it = get_global_context_value("integral_type")
    if it == "cell":
        subelem_inames = sub_element_inames()
    elif it == "exterior_facet" or it == "interior_facet":
        subelem_inames = sub_facet_inames()

    k = get_option("number_of_blocks")
    modified_index = prim.Sum((tensor_index_to_sequential_index(to_tensor_index(iname), k+1),
                               tensor_index_to_sequential_index(tuple(prim.Variable(iname) for iname in subelem_inames), k + 1)))

    return modified_index

